Package: JuliaConnectoR
Type: Package
Title: A Functionally Oriented Interface for Integrating 'Julia' with R
Version: 1.1.3
Maintainer: Stefan Lenz <stefan-m-lenz@web.de>
Authors@R: c(person("Stefan", "Lenz", role = c("aut", "cre"),
                    email = "stefan-m-lenz@web.de",
                    comment = c(ORCID = "0000-0001-9135-1743")),
             person("Harald", "Binder", role = c("aut", "ths"),
                    comment = c(ORCID = "0000-0002-5666-8662")),
             person("Angelo", "D\'Ambrosio",
                    comment = c(ORCID = "0000-0002-2045-5155"), role = "ctb"))
Description: Allows to import functions and whole packages from 'Julia' in R.
    Imported 'Julia' functions can directly be called as R functions.
    Data structures can be translated between 'Julia' and R.
    More details can also be found in the corresponding article
    <doi:10.18637/jss.v101.i06>.
License: MIT + file LICENCE
SystemRequirements: Julia >= 1.0
Encoding: UTF-8
RoxygenNote: 7.2.3
Suggests: testthat (>= 2.1.0), future, parallel
NeedsCompilation: no
Packaged: 2023-12-06 13:57:28 UTC; lenzstef
Author: Stefan Lenz [aut, cre] (<https://orcid.org/0000-0001-9135-1743>),
  Harald Binder [aut, ths] (<https://orcid.org/0000-0002-5666-8662>),
  Angelo D'Ambrosio [ctb] (<https://orcid.org/0000-0002-2045-5155>)
Repository: CRAN
Date/Publication: 2023-12-06 15:30:02 UTC
Built: R 4.4.0; ; 2024-05-01 13:31:33 UTC; unix
